local MODNAME, ShuckIt = ...

-- ShuckIt.lua
-- $Id: ShuckIt.lua 62 2013-12-26 18:11:36Z LaoTseu $

-- The ShuckIt idea came from the Shucks module from Cork (http://www.wowinterface.com/downloads/info10629) by Tekkub.
-- The basic framework for this addon is based on Tekkub's AddonTemplate (http://github.com/tekkub/addontemplate/tree/master)
-- The code to find the openable items is based on Ammo's work in
-- CrowBar (http://www.wowace.com/addons/crow-bar/) and is used with his permission.


---------------------------
-- Localized Lua globals --
---------------------------

local _G = getfenv(0)

local assert = _G.assert
local error = _G.error
local geterrorhandler = _G.geterrorhandler
local ipairs = _G.ipairs
local pairs = _G.pairs
local rawset = _G.rawset
local select = _G.select
local setmetatable = _G.setmetatable
local sort = _G.sort
local strjoin = _G.strjoin
local tinsert = _G.tinsert
local tonumber = _G.tonumber
local tostring = _G.tostring
local tremove = _G.tremove

----------------------
--     Revision     --
----------------------


ShuckIt = ShuckIt or {}
ShuckIt.revision = ShuckIt.revision or {}
ShuckIt.revision.main	= ("$Revision: 62 $"):match("(%d+)")
ShuckIt.version      	= _G.GetAddOnMetadata("ShuckIt", "Version"):match("([^ ]+)")
ShuckIt.revision.toc  	= _G.GetAddOnMetadata("ShuckIt", "Version"):match("%$Revision:%s(%d+)")

----------------------
--      Locals      --
----------------------

local L = ShuckIt.localized
--local defaults, defaultsPC, db, dbpc = {}, {}
local UPDATEPERIOD, elapsed = 0.5, 0
local TOOLTIPPERIOD, tooltip_elapsed = 0.2, 0
local need_scan, need_chat_update, busy, delay_scan = nil, nil, nil, nil

local item_to_shuck, icon, shuck_count = nil, nil, 0
-- I'm creating three tables because I don't want to deal with recycling sub-table
-- The ideal data type would have been a table of triplets { itemID, bag, slot }
local shuck_table = {}
local shuck_table_bag = {}
local shuck_table_slot = {}
local shuck_bag, shuck_slot = 1, 1
local default_icon = "Interface\\Icons\\INV_Misc_Shell_03"
local found_quest_item
local HEROIC_OPEN = L["Use: Open the Heroic Cache of Treasures to earn your glorious reward! (1 Sec Cooldown)"]

-- Frames
local f
local secureframe

local lqt = _G.LibStub("LibQTip-1.0")
local ldb = _G.LibStub:GetLibrary("LibDataBroker-1.1")
local dataobj = ldb:NewDataObject("ShuckIt",
											 {
												type = "data source",
											  	text = "",
											  	icon = default_icon
})

-- Special containers: containers that are not clams but like clams do not have the <right click to open> text
local special_containers = {
	[58856]		= true,	-- Royal Monkfish
}

------------------------------
--      Util Functions      --
------------------------------

local function err(msg,...) _G.geterrorhandler()(msg:format(_G.tostringall(...)) .. " - " .. _G.time()) end

local default_channel = nil
local function setDefaultChanelForPrint()
	default_channel = nil
	for i=1, _G.NUM_CHAT_WINDOWS do
		local name = _G.GetChatWindowInfo(i)
		if name and name:lower() == 'output' then
			default_channel = i
		end
	end
end

local function Print(message, ...)
	if default_channel then
		_G["ChatFrame"..default_channel]:AddMessage(("|cff00dbba"..L["ShuckIt"].."|r: "..message):format(...));
	else
		_G.SELECTED_CHAT_FRAME:AddMessage(("|cff00dbba"..L["ShuckIt"].."|r: "..message):format(...))
	end
end

local debugf = _G.tekDebug and _G.tekDebug:GetFrame("ShuckIt")
local function Debug(...) if debugf then debugf:AddMessage(strjoin(", ", ...)) end end

------------------------------
--      Core Functions     --
------------------------------

local ShuckItTooltip = _G.CreateFrame("GameTooltip", "ShuckItTooltip")
ShuckItTooltip:SetOwner(_G.WorldFrame, "ANCHOR_NONE")

ShuckItTooltip:AddFontStrings(
    ShuckItTooltip:CreateFontString( "$parentTextLeft1", nil, "GameTooltipText" ),
    ShuckItTooltip:CreateFontString( "$parentTextRight1", nil, "GameTooltipText" )
)

local cache = {}
local do_not_cache = {}
local function IsOpenable(bag, slot, itemLink)


	if not bag or not slot then return false end
	if not itemLink then itemLink = _G.GetContainerItemLink(bag, slot) end
	if not itemLink then return false end -- Bag slot empty

	-- Is this a quest item for a quest that has not been activated yet ?
	local _, questID, isActive = _G.GetContainerItemQuestInfo(bag, slot)
	if cache[itemLink] ~= nil then return cache[itemLink] end

	if questID then
--err("questID = %s, isActive = %s, bag = %s, slot = %s, itemLink = %s", questID, isActive, bag, slot, itemLink)	
		local minLevel = select(5, _G.GetItemInfo(itemLink))
		if minLevel <= _G.UnitLevel("player") then
			found_quest_item = true
			-- No caching since a quest can be abandoned
			if not isActive and not _G.IsQuestFlaggedCompleted(questID) then
				return true
			else
				return false
			end
		end
	end


	-- Tooltip parsing
	ShuckItTooltip:ClearLines()
	if not ShuckItTooltip:IsOwned(_G.WorldFrame) then ShuckItTooltip:SetOwner(_G.WorldFrame, "ANCHOR_NONE") end
 	ShuckItTooltip:SetBagItem(bag,slot)

	for i=1,ShuckItTooltip:NumLines() do
	   local mytext = _G["ShuckItTooltipTextLeft" .. i]
	   local text = mytext:GetText()

		if (text == _G.ITEM_OPENABLE) then
			if not do_not_cache[itemLink] then cache[itemLink] = true end
			return true
		elseif text == _G.LOCKED then
			-- Do not cache the result for locked items, they may become unlocked
			do_not_cache[itemLink] = true
			return false
		end
	end

	-- Is the item lootable according to GetContainerItemInfo?
	-- This needs to be done after the tooltip parsing otherwise the locked items are not detected.
	--  F98D00 (unlocked) 6B5B900 (locked)
	local lootable = select(6, _G.GetContainerItemInfo(bag, slot))
	if lootable then
		if not do_not_cache[itemLink] then cache[itemLink] = true end
		return true
	end

	cache[itemLink] = false
	return false

end

_G.IsOpenable = IsOpenable

-- Cache some items that cannot be detected. If anyone has a clue how to detect that these are openable, please let me know
-- These work like clams in the sence that UseContainerItem will not work on them
local knownClams = {
	[89125]	= true,		-- Sack of pet supplies
	[91086]	= true,		-- Darkmoon Pet Supplies
	[93146]	= true,		-- Pandaren Spirit Pet Supplies
	[93147]	= true,		-- Pandaren Spirit Pet Supplies
	[93148]	= true,		-- Pandaren Spirit Pet Supplies
	[93149]	= true,		-- Pandaren Spirit Pet Supplies
	[94207]	= true,		-- Fabled Pandaren Pet Supplies
	[98095]	= true,		-- Brawler's Pet
}
local cache_clam = {}
local function IsAClam(bag, slot, itemLink)
	if not bag or not slot then return false end
	if not itemLink then itemLink = _G.GetContainerItemLink(bag, slot) end
	if not itemLink then return false end -- Bag slot empty

	-- Items that we cannot detect but that are openable
	local itemID = itemLink:match("item:(%d+)")
	if knownClams[tonumber(itemID)] then return true end

	if cache_clam[itemLink] ~= nil then return cache_clam[itemLink] end

	ShuckItTooltip:ClearLines()
	if not ShuckItTooltip:IsOwned(_G.WorldFrame) then ShuckItTooltip:SetOwner(_G.WorldFrame, "ANCHOR_NONE") end
 	ShuckItTooltip:SetBagItem(bag,slot)

	for i=1,ShuckItTooltip:NumLines() do
	   local mytext = _G["ShuckItTooltipTextLeft" .. i]
	   local text = mytext:GetText()

		if (text == L["Use: Open the clam!"] or text == HEROIC_OPEN) then
			cache_clam[itemLink] = true
			return true
		end
	end

	cache_clam[itemLink] = false
	return false

end

local function ScanBags()

	item_to_shuck, shuck_bag, shuck_slot = nil, nil, nil
	shuck_count = 0
	found_quest_item = nil
	for k in pairs(shuck_table) do shuck_table[k] = nil end
	for k in pairs(shuck_table_bag) do shuck_table_bag[k] = nil end
	for k in pairs(shuck_table_slot) do shuck_table_slot[k] = nil end

	local newIcon = nil
	for bag = 0, _G.NUM_BAG_FRAMES, 1 do
		for slot = 1, _G.GetContainerNumSlots(bag), 1 do

			local itemLink = _G.GetContainerItemLink(bag, slot)
			if itemLink and (IsOpenable(bag, slot, itemLink) or IsAClam(bag, slot, itemLink)) then
				local itemID = itemLink:match("item:(%d+)")
				shuck_table[itemID] = (shuck_table[itemID] or 0) + tonumber(select(2,_G.GetContainerItemInfo(bag, slot)) or 1)
				-- we keep the bag and slot for the first item we find
				shuck_table_bag[itemID] = shuck_table_bag[itemID] or bag
				shuck_table_slot[itemID] = shuck_table_slot[itemID] or slot
				if not item_to_shuck then
					item_to_shuck = _G.GetItemInfo(itemLink)
					shuck_bag, shuck_slot = bag, slot
					newIcon = _G.GetContainerItemInfo(bag, slot)
				end
			elseif itemLink and not _G.GetItemInfo(itemLink) then
				-- This is a workaround for a intermitent situation when opening
				-- multiple items of the same kind in a row sometimes result in
				-- the item information being screewed up.
				-- If this happen, we delay de scan until the information is realiable
				delay_scan = true
				return
			end

		end
	end

	-- If there is a quest item, we must listen to quest log events
	if found_quest_item then
--err("found_quest_item = %s", found_quest_item)	
		f:RegisterEvent("UNIT_QUEST_LOG_CHANGED")
	else
		f:UnregisterEvent("UNIT_QUEST_LOG_CHANGED")
	end

	for k in pairs (shuck_table) do
		shuck_count = shuck_count + shuck_table[k]
	end
	icon = newIcon or default_icon
	need_scan = nil
end

local function myOnClick(self, button, down)
	if not _G.InCombatLockdown() and not _G.IsStealthed() and not busy and IsOpenable(shuck_bag, shuck_slot) then
		_G.UseContainerItem(shuck_bag, shuck_slot)
		return true
	elseif not _G.InCombatLockdown() and not _G.IsStealthed() and not busy and IsAClam(shuck_bag, shuck_slot) then
		Print(L["Cannot open Clams by clicking on the LDB display. Use %s instead."],"|cffffd700/click ShuckItFrame|r")
	else
		return false
	end
end

-----------------------------
--      Event Handler      --
-----------------------------

f = _G.CreateFrame("frame")
f:SetScript("OnEvent", function(self, event, ...) if self[event] then return self[event](self, event, ...) end end)
f:RegisterEvent("ADDON_LOADED")


function f:ADDON_LOADED(event, addon)
	if addon ~= "ShuckIt" then return end

	self:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil

	if _G.IsLoggedIn() then self:PLAYER_LOGIN() else self:RegisterEvent("PLAYER_LOGIN") end
end


function f:PLAYER_LOGIN()
	-- Create the secure frame
	ShuckIt:InitSecureFrame()

	-- Register events
	self:RegisterEvent("BAG_UPDATE")
	self:RegisterEvent("UNIT_SPELLCAST_SENT")
	self:RegisterEvent("UNIT_SPELLCAST_FAILED")
	self:RegisterEvent("UNIT_SPELLCAST_STOP")
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("UPDATE_CHAT_WINDOWS")

	-- Start the OnUpdate
	self:SetScript("OnUpdate", function(self, elap)
		elapsed = elapsed + elap
		tooltip_elapsed = tooltip_elapsed + elap

		-- Tooltip management
		if tooltip_elapsed >= TOOLTIPPERIOD then
			tooltip_elapsed = 0
			if not (ShuckIt.tooltip and ShuckIt.tooltip:IsMouseOver()) and
				not (ShuckIt.tooltip_anchor and ShuckIt.tooltip_anchor:IsMouseOver()) then
				ShuckIt.tooltip = lqt:Release(ShuckIt.tooltip)
				ShuckIt.tooltip_anchor = nil
			end
		end

		if delay_scan then
			-- Workaround for when good item information is not available and we
			-- need to wait a bit
			elapsed = UPDATEPERIOD/2
			need_scan = true
			delay_scan = nil
			return
		end

		if not (need_scan or need_chat_update) or busy or elapsed < UPDATEPERIOD or _G.InCombatLockdown() then return end


		elapsed = 0

		-- Print window if needed
		if need_chat_update then
			setDefaultChanelForPrint()
			need_chat_update = nil
		end

		-- Find object to Shuck
		ScanBags()

		-- Adjust the display
		if shuck_count > 0 then
			dataobj.text = (L["%s - %d"]):format(item_to_shuck, shuck_count)
		else
			dataobj.text = ""
		end
		dataobj.icon = icon

		-- Update the Tooltip
		if ShuckIt.tooltip then ShuckIt:DrawTooltip() end
	end)

	-- Set the OnClick function for the data broker
	dataobj.OnClick = myOnClick

	-- Find the revision number
	if ShuckIt.revision then
		ShuckIt.rev = 0
		for _, rev in pairs(ShuckIt.revision) do
			local revision = tonumber(rev)
			if revision and revision > ShuckIt.rev then ShuckIt.rev = revision end
		end
	else
		assert(false,"No ShuckIt.revision")
	end

	-- Find the default channel
	setDefaultChanelForPrint()

	need_scan = true

	self:UnregisterEvent("PLAYER_LOGIN")
	self.PLAYER_LOGIN = nil
end


function f:BAG_UPDATE(message, bagID)
--err("BAG_UPDATE")
	elapsed = 0 -- Let's wait a bit
	need_scan = true
end

function f:UNIT_QUEST_LOG_CHANGED(message, unit)
--err("UNIT_QUEST_LOG_CHANGED: message=%s, unit=%s", message, unit)
	if unit == "player" then
		elapsed = 0 -- Let's wait a bit
		need_scan = true
	end
end

function f:UNIT_SPELLCAST_SENT(event, unit, spell, rank, target)
	busy = true
end

function f:UNIT_SPELLCAST_FAILED(event, unit, spell, rank)
	busy = nil
end

function f:UNIT_SPELLCAST_STOP(event, unit, spell, rank)
	busy = nil
end

function f:UNIT_SPELLCAST_INTERRUPTED(event, unit, spell, rank)
	busy = nil
end

local lock_picking_spell = _G.GetSpellInfo(1804)
function f:UNIT_SPELLCAST_SUCCEEDED(event, unit, spell, rank)
	busy = nil
	if unit == "player" and spell == lock_picking_spell then
		-- A something just got unlocked, it can be opened
		elapsed = -UPDATEPERIOD -- Let's wait a bit
		need_scan = true
		--Print("Something was unlocked")
	end
end

function f:UPDATE_CHAT_WINDOWS()
	need_chat_update = true
end


-----------------------------
--      Slash Handler      --
-----------------------------

_G.SLASH_SHUCKIT1 = "/shuckit"
_G.SlashCmdList.SHUCKIT = function(msg)
	Print(L["Use %s to open the next item"],"|cffffd700/click ShuckItFrame|r")
end

-----------------------------
--        LDB Stuff        --
-----------------------------

-- Sort function
local function sort_by_item_name (a,b)
	local name_a = _G.GetItemInfo(a)
	local name_b = _G.GetItemInfo(b)
	return name_a < name_b
end


local index = {}
function ShuckIt:DrawTooltip(anchor)
	self.tooltip_anchor = anchor or self.tooltip_anchor

	if not self.tooltip then
		self.tooltip = lqt:Acquire("ShuckItTooltip", 1, "LEFT")
	end
	local tooltip = self.tooltip

	tooltip:Clear()
	tooltip:SmartAnchorTo(self.tooltip_anchor)
	tooltip:SetScale(1)

	local line, column = tooltip:AddHeader()
	tooltip:SetCell(line, 1, ("|cffffd700%s v%s (r%d)|r"):format(L["ShuckIt"], ShuckIt.version, ShuckIt.rev), "CENTER", 1)

	-- clean index
	for key in pairs(index) do index[key] = nil end

	-- Sort by item name
	for key in pairs(shuck_table) do tinsert(index,key) end
	sort(index, sort_by_item_name)

	if #index > 0 then
		line, column = tooltip:AddHeader()
		local item = L["item"]
		if shuck_count > 1 then item = L["items"] end
		tooltip:SetCell(line, 1, (L["|cffffd700%d %s to open|r"]):format(shuck_count,item), "CENTER", 1)
		tooltip:AddSeparator()
	end

	for _, link in ipairs (index) do
		local count = shuck_table[link]
		local name, _, _, _, _, _, _, _, _, texture = _G.GetItemInfo(link)

		line, column = tooltip:AddLine()
		tooltip:SetCell(line, 1, (L["|T%s:16|t %s x %d"]):format(
												  texture,
												  name,
												  count
										 ), "LEFT")
		tooltip:SetCellScript(line, column, "OnMouseUp", function()
			if not _G.InCombatLockdown() and not _G.IsStealthed() and not busy and IsOpenable(shuck_table_bag[link], shuck_table_slot[link]) then
				_G.UseContainerItem(shuck_table_bag[link], shuck_table_slot[link])
				return true
			elseif not _G.InCombatLockdown() and not _G.IsStealthed() and not busy and IsAClam(shuck_table_bag[link], shuck_table_slot[link]) then
				Print(L["Cannot open Clams by clicking on the LDB display. Use %s instead."], "|cffffd700/click ShuckItFrame|r")
			end
 		end)
	end

	tooltip:Show()

end

function dataobj:OnEnter(motion)
	ShuckIt:DrawTooltip(self)
	tooltip_elapsed = 0
end

function dataobj:OnLeave()
   --lqt:Release(ShuckIt.tooltip)
   --ShuckIt.tooltip = nil
end

----------------------------
--      Secure frame      --
----------------------------

do

	local frame = nil
	function ShuckIt:InitSecureFrame()
		--err("ShuckIt:InitSecureFrame")
		if _G.InCombatLockdown() or frame then

			if not frame then

				frame = _G.CreateFrame("frame","SkuckIt_InitSecureFrame")
				assert(frame, "Was not able to create frame")

				frame:Hide()

				frame:SetScript("OnUpdate", function(self, elap)
						if not _G.InCombatLockdown then
							ShuckIt:InitSecureFrame()
							frame:Hide()
							frame = nil
						end
				end)
				frame:Show()
			end

			return
		end

		secureframe = _G.CreateFrame("Button", "ShuckItFrame", _G.UIParent, "SecureActionButtonTemplate")

		secureframe:SetScript("PreClick", function(self)
			if _G.InCombatLockdown() or _G.IsStealthed() or busy or
					(not IsOpenable(shuck_bag, shuck_slot) and not IsAClam(shuck_bag, shuck_slot)) then
				Print(L["Nothing to do or in combat"])
				return
			end
			self:SetAttribute( "type1", "macro" )
			self:SetAttribute( "macrotext1", "/use "..shuck_bag.." "..shuck_slot)
		end)


		secureframe:SetScript("PostClick", function(self)
			if _G.InCombatLockdown() then return end
			self:SetAttribute( "type1", _G.ATTRIBUTE_NOOP )
			self:SetAttribute( "macrotext1", nil)
		end)
		--err("Frame created: %s",secureframe)
	end

end